"use strict";

/**
 * @typedef {object} AlertBox
 * @property {function(type: string, message: string, category: string, hide_previous: boolean)} show Show alert message
 * @property {function(message: string, category: string = null, hide_previous: boolean = true)} showError
 * @property {function(message: string, category: string = null, hide_previous: boolean = true)} showWarn
 * @property {function(message: string, category: string = null, hide_previous: boolean = true)} showSuccess
 * @property {function(message: string, category: string = null, hide_previous: boolean = true)} showInfo
 * @property {function(category: string)} hide Hide all messages in category
 * @property {function()} clear Hide all alerts
 */

/**
 * Display alert messages in given container
 * If message won't fit into viewport (scrolled), it will display in hover_container (optional)
 *
 * @param {jQuery} container
 * @param {jQuery|null} [hover_container = null]
 * @return {AlertBox}
 * @constructor
 */
function AlertBox(container, hover_container = null) {

  /**
   * @param {string} type
   * @param {string} message
   * @param {string} category
   * @param {boolean} hide_previous_in_category Hide previous message(s) in category. Default is true.
   */
  function show(type, message, category, hide_previous_in_category = true) {
    if (category && hide_previous_in_category) {
      hide(category);
    }
    const id = "alert_" + String(Math.floor(Math.random() * 1000000));
    const html = getAlertMessageHtml(type, message, category, id);
    const alertsBoundingBox = container[0].getBoundingClientRect();

    if (hover_container === null || alertsBoundingBox.bottom + 25 > 0) { // 25px - height of one-line alert message
      container.append(html);
    } else {
      hover_container.append(html);
      setTimeout(function () {
        const alert = hover_container.find('.' + id);
        alert.hide(250, function () {
          alert.detach();
          container.append(alert.show());
        });
      }, 3000);
    }
  }

  function showError(message, category = '', hide_previous_in_category = true) {
    show('error', message, category, hide_previous_in_category);
  }

  function showWarn(message, category = '', hide_previous_in_category = true) {
    show('warning', message, category, hide_previous_in_category);
  }

  function showSuccess(message, category = '', hide_previous_in_category = true) {
    show('success', message, category, hide_previous_in_category);
  }

  function showInfo(message, category = '', hide_previous_in_category = true) {
    show('info', message, category, hide_previous_in_category);
  }


  function getAlertMessageHtml(type, message, category, id) {
    var strongmess='';
    if (type === "error") { type='danger'; }
    if (type === "danger") { strongmess='Error'; }
    if (type === "warning") { strongmess='Warning'; }
    if (type === "success") { strongmess='Success'; }
    if (type === "info") { strongmess='Info'; }

    return `<div category="${category || ''}" class="w-100 alert alert-${type} alert-dismissible ${id}"><button type="button" class="close" data-dismiss="alert" aria-hidden="true">&times;</button><strong>${strongmess}: </strong> ${message}</div>`;
  }

  function hide(category = '') {
    container.find('div[category="' + category + '"]').remove();
    if (hover_container !== null) {
      hover_container.find('div[category="' + category + '"]').remove();
    }
  }

  function clear() {
    container.children('div.alert').remove();
  }

  if (container.length === 0) {
    console.error('AlertBox initialized with non-existing container');
  }

  return Object.freeze({
    show,
    showError,
    showWarn,
    showSuccess,
    showInfo,
    hide,
    clear,

  });
}
