'use strict';

let Camera = function() {

  let getParameter = async function(name) {
    let parameters = await this.getParameters([name]);
    return parameters[name];
  };

  let getParameters = async function(parameters) {
    if (typeof parameters !== 'object' || parameters === null) {
      throw new Error("Argument must be an object or Array");
    }

    if (Array.isArray(parameters)) {
      parameters = Object.fromEntries(parameters.map( name => [name, name] ));
    }

    let param_names = Object.values(parameters);
    if (param_names.some( name => typeof name === "string" && name.match(/\.StreamList/i)) && param_names.length !== 1) {
      throw new Error("Can't read parameters containing newlines together with another parameter.");
    }

    let group = param_names.join(',');
    let params = {
      dummy: new Date().getTime(),
      action: 'list',
      group: group
    };

    let response = await callApi("/axis-cgi/param.cgi", params);

    if (typeof response !== 'string' || response === '') {
      throw new Error('Empty response from camera');
    }

    //Parse response
    let param_values = {};
    if (param_names.length === 1) {  //response for e.g. StreamList contains newline characters and would break logic using String.split("\n").
      let equal_sign_position = response.indexOf('=');
      if (equal_sign_position === -1) {
        return [];
      }
      param_values[response.substr(0, equal_sign_position)] = response.substr(equal_sign_position + 1).trim(); //trim the newline char
    } else {
      let splitted_params = response.split("\n");
      for (let i = 0; i < splitted_params.length; i++) {
        if (splitted_params[i].indexOf('=') === -1) {
          continue;
        }
        let splitted_value = splitted_params[i].split('=');
        param_values[splitted_value.shift()] = splitted_value.join('=');
      }
    }

    // Prepare output object
    let renamed_param_values = {};
    for (let key in parameters) {
      if (param_values.hasOwnProperty(parameters[key])) {
        renamed_param_values[key] = param_values[parameters[key]];
      }
    }

    return renamed_param_values;
  };

  /**
   * Call Axis Api with Fetch
   * @param {string} endpoint
   * @param {object} params
   */
  let callApi = async function(endpoint, params) {
    let request = new Request(endpoint, {
      method: 'POST',
      headers: {
        'Content-Type': 'application/x-www-form-urlencoded',
      },
      body: buildQueryString(params),
    });

    let response;
    try {
      response = await fetch(request);
      if (!response.ok) {
        throw Error('Communication error');
      }
      response = await response.text();
    } catch (e) {
      if (!e.status) {
        throw new Error('Network error');
      }
      throw new Error(`HTTP ${e.status}: ${e.statusText}`);
    }
    return response;
  }

  /**
   * @param {object} parameters
   * @return {Promise<void>}
   */
  let setParameters = async function(parameters) {
    if (typeof parameters !== 'object' || Array.isArray(parameters) || parameters === null) {
      throw new Error("Argument must be an object");
    }

    let query_params = parameters;
    query_params['action'] = 'update';

    let response = await callApi("/axis-cgi/param.cgi", query_params);

    if (response !== 'OK') {
      throw new Error('Error setting parameter');
    }
  };

  /**
   * @param {string} mode [text, zip, zip_with_image]
   * @return {Promise<Blob>}
   */
  let getServerReport = async function(mode = 'text') {
    try {
      const url = '/axis-cgi/admin/serverreport.cgi?mode=' + mode;
      const response = await fetch(url);
      if (!response.ok) {
        throw new Error('HTTP error, status = ' + response.status);
      }

      return response.blob();
    } catch (e) {
      throw new Error(`Failed to fetch server report. ${e.message}`);
    }
  };

  /**
   * @param {Array.<{name: String, data: Blob}>} attachments
   * @return {Promise<Blob>}
   */
  let getReport = async function(attachments = []) {

    let data = new FormData();
    attachments.forEach( item => data.set(item.name, item.data));

    let request = new Request("/local/camstreamer/report.cgi", {
      method: 'POST',
      body: data,
      cache: "no-cache",
    });

    let response;
    try {
      response = await fetch(request);
    } catch (e) {
      throw new Error(`Failed to fetch report. ${e.message}`);
    }

    const content_type = response.headers.get('Content-Type');
    if (content_type === 'application/zip') {
      return response.blob();
    } else if (content_type === 'application/json') {
      const json = response.json();
      throw new Error(json.message || json.status || 'Invalid response from camera.');
    } else {
      throw new Error(`Invalid response content type '${content_type}'.`)
    }
  };

  /**
   * Get max fps for channel using capturemode.cgi
   * Supported since firmware 8.50
   * @param {number} channel_id
   * @return {Promise<number>}
   */
  let getMaxFps = async function(channel_id)
  {
    let request = new Request("/axis-cgi/capturemode.cgi", {
      method: 'POST',
      body: '{"apiVersion": "1.0","context": "abc","method": "getCaptureModes"}',
      cache: "no-cache",
    });

    let response;
    try {
      const reply = await fetch(request);
      response = await reply.json();
    } catch (e) {
      throw new Error(`Failed to fetch report. ${e.message}`);
    }

    /** @var {Array} */
    const channels = response.data;
    if (channels === undefined) {
      throw new Error(`Malformed reply from camera`);
    }
    const channel = channels.find(x => x.channel === channel_id);

    if (channel === undefined) {
      throw new Error(`Video channel '${channel_id}' not found`);
    }

    /** @var {Array} */
    const capture_modes = channel.captureMode;
    const capture_mode = capture_modes.find(x => x.enabled === true);
    if (capture_mode === undefined) {
      throw new Error(`No enabled capture mode found.`);
    }

    const max_fps = parseInt(capture_mode.maxFPS, 10);
    if (isNaN(max_fps)) {
      throw new Error(`Max fps not specified for given capture mode.`);
    }

    return max_fps;
  };

  return Object.freeze({
    getParameter,
    getParameters,
    setParameters,
    getServerReport,
    getReport,
    getMaxFps,
  });
};