'use strict';

/**
 * @typedef {object} WsClient
 * @property {function()} sendInitData Request init_data from camera
 * @property {function()} close Close connection to ACAP without reconnecting.
 * @property {function(Window, string)} addPushMessageTarget Register new recipient for WS messages and events
 */

/**
 * @param {string} host
 * @param {boolean} secure
 * @return {WsClient}
 */
function websocketClient(host, secure) {
  'use strict';

  const protocol = secure ? 'wss' : 'ws';
  const url = `${protocol}://${host}/local/camstreamer/events`;

  /** @type {WebSocket} */
  let websocket;

  /**
   * Registered handlers for incoming WS messages
   * @type {Array<function(Object)>}
   */
  const event_listeners = [];

  /**
   * Registered handlers for WS connection state changes
   * @type {Array<function(Object)>}
   */
  const state_listeners = [];

  let is_socket_authorized = false;

  let authorization_timeout_id;

  let retry_counter = 0;

  function init() {
    websocket = new WebSocket(url, 'events');

    websocket.onopen = handleSocketOpen;
    websocket.onclose = handleSocketClose;
    websocket.onmessage = handleSocketMessage;
    websocket.onerror = handleSocketError;

    clearTimeout(authorization_timeout_id);
    authorization_timeout_id = setTimeout(function () {
      console.warn(`Websocket authorization timeout. Reconnecting...`);
      websocket.close(4002);
    }, 10000);
  }

  async function handleSocketOpen() {
    retry_counter = 0;

    let token;
    try {
      token = await getAuthToken();
    } catch (err) {
      console.warn(`Failed to get websocket auth token (${err.message}). Reconnecting...`);
      websocket.close(4003);
      return;
    }

    websocket.send(JSON.stringify({authorization: token}));
  }

  /**
   * @param {number} retry_counter
   * @return {number} msec
   */
  function countReconnectTimeout(retry_counter) {
    const data = [
      [10, 30000],
      [ 6,  5000],
      [ 3,  2000],
      [ 0,  1000],
    ];
    for (const row of data) {
      if (retry_counter > row[0]) {
        return row[1];
      }
    }

    return 1000;
  }

  function handleSocketClose(evt) {
    const no_reconnect_codes = [ 1000 ];
    is_socket_authorized = false;

    clearTimeout(authorization_timeout_id);

    if (!no_reconnect_codes.includes(evt.code)) {
      retry_counter++;

      const reconnect_timeout = countReconnectTimeout(retry_counter);
      const reconnect_seconds = Math.round(reconnect_timeout / 1000);
      console.warn(`Camera WebSocket closed. Will try to reconnect in ${reconnect_seconds}s.`, evt.code);
      reconnect(reconnect_timeout);

      state_listeners.forEach(function (callback) {
        callback(WebSocket.CLOSED, {code: evt.code});
      });
    }
  }

  function handleSocketMessage(evt) {
    let data;
    try {
      data = JSON.parse(evt.data);
    } catch (e) {
      console.error('Invalid JSON in notification message: ' + evt.data);
      return;
    }
    if( !data.type || typeof data.type !== 'string') {
      console.error('No message type given.');
      return;
    }

    if (data.type === 'authorization') {
      handleSocketAuthResponse(data.state);
      return;
    }

    event_listeners.forEach(function (callback) {
      callback(data);
    });
  }

  function handleSocketError(evt) {
    console.error("Camera WebSocket error. Will try to reconnect.", evt);
    websocket.close(4000);
  }

  /**
   * @return {Promise<string>}
   */
  async function getAuthToken() {
    let response = await $.get('/local/camstreamer/ws_authorization.cgi')
      .catch( (jqXHR, textStatus, errorThrown) => { throw new Error(errorThrown || 'Network error') });

    if (response.status !== 200) {
      throw new Error(response.message || 'Bad response.');
    }
    return response.data;
  }

  function handleSocketAuthResponse(auth_state) {
    clearTimeout(authorization_timeout_id);
    if (auth_state !== 'OK') {
      console.warn(`Socket authorization failed (${auth_state}). Reconnecting...`);
      websocket.close(4001);
    }
    if (!is_socket_authorized) {
      is_socket_authorized = true;
      state_listeners.forEach(function (callback) {
        callback(WebSocket.OPEN, {});
      });
    } else {
      console.warn('handleSocketAuthResponse called while socket is already in authorized state');
    }
  }

  let onEvent = function (callback) {
    event_listeners.push(callback);
  };

  let onStateChange = function (callback) {
    state_listeners.push(callback);
  };

  let reconnectTimeout = null;

  /**
   * @param {number} timeout msec
   */
  let reconnect = function(timeout = 100) {
    if (websocket && [ 0, 1 ].includes(websocket.readyState)) {
      websocket.close(1000);
    }
    if (!reconnectTimeout) {
      reconnectTimeout = setTimeout(function () {
        init();
        reconnectTimeout = null;
      }, timeout);
    }
  };

  let sendInitData = function() {
    if (websocket.readyState === WebSocket.CONNECTING || !is_socket_authorized) {
      return; //init_data are sent after successful authorization
    }
    websocket.send(JSON.stringify({command: 'sendInitData'}));
  };

  let close = function () {
    websocket.close(1000);
  };

  let addPushMessageTarget = function (target_window, target_origin) {

    onEvent(function (event) {
      const message = {
        query: 'WebsocketEvent',
        data: event
      };
      target_window.postMessage(message, target_origin);
    });

    onStateChange(function (readyState, state_data) {
      const message = {
        query: 'WebsocketStateChange',
        data: {
          state: readyState,
          data: state_data,
        }
      };
      target_window.postMessage(message, target_origin);
    });
  };

  init();

  return Object.freeze({
    sendInitData,
    close,
    addPushMessageTarget,
  });
}